<?php

declare(strict_types=1);

namespace MsgPhp\Domain\Infra\DependencyInjection;

use Symfony\Component\DependencyInjection\ChildDefinition;
use Symfony\Component\DependencyInjection\ContainerBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\DependencyInjection\Definition;
use Symfony\Component\DependencyInjection\Exception\InvalidArgumentException;

/**
 * @author Roland Franssen <franssen.roland@gmail.com>
 *
 * @internal
 */
final class ContainerHelper
{
    private static $counter = 0;

    public static function hasBundle(ContainerInterface $container, string $class): bool
    {
        return in_array($class, $container->getParameter('kernel.bundles'), true);
    }

    public static function getBundles(ContainerInterface $container): array
    {
        return array_flip($container->getParameter('kernel.bundles'));
    }

    public static function getClassReflection(ContainerBuilder $container, ?string $class): \ReflectionClass
    {
        if (!$class || !($reflection = $container->getReflectionClass($class))) {
            throw new InvalidArgumentException(sprintf('Invalid class "%s".', $class));
        }

        return $reflection;
    }

    public static function removeId(ContainerBuilder $container, string $id): void
    {
        $container->removeDefinition($id);
        $container->removeAlias($id);

        foreach ($container->getAliases() as $aliasId => $alias) {
            if ($id === (string) $alias) {
                $container->removeAlias($aliasId);
            }
        }
    }

    public static function removeIf(ContainerBuilder $container, bool $condition, array $ids): void
    {
        if (!$condition) {
            return;
        }

        foreach ($ids as $id) {
            self::removeId($container, $id);
        }
    }

    public static function registerAnonymous(ContainerBuilder $container, string $class, bool $child = false, string &$id = null): Definition
    {
        $definition = $child ? new ChildDefinition($class) : new Definition($class);
        $definition->setPublic(false);

        return $container->setDefinition($id = $class.'.'.ContainerBuilder::hash(__METHOD__.++self::$counter), $definition);
    }

    public static function tagCommandHandler(ContainerBuilder $container, string $handlerId, array $handles, string $busId): void
    {
        $handler = $container->getDefinition($handlerId);
        $messengerEnabled = FeatureDetection::isMessengerAvailable($container);
        $simpleBusEnabled = FeatureDetection::hasSimpleBusCommandBusBundle($container);

        foreach ($handles as $class) {
            if ($messengerEnabled) {
                $handler->addTag('messenger.message_handler', ['handles' => $class, 'bus' => $busId]);
            }
            if ($simpleBusEnabled) {
                $handler
                    ->setPublic(true)
                    ->addTag('command_handler', ['handles' => $class]);
            }
        }
    }

    private function __construct()
    {
    }
}
